"""End to End System test"""

import argparse
import contextlib
import http.client
import json
import logging
import sys
import warnings
from datetime import datetime, timedelta
from typing import Any, List, Optional

import click
import requests
from fido2.hid import CtapHidDevice
from requests import Response
from urllib3.exceptions import InsecureRequestWarning

from pyicloud import PyiCloudService
from pyicloud.exceptions import PyiCloudServiceUnavailable
from pyicloud.services.calendar import CalendarObject, CalendarService

END_LIST = "End List\n"
MAX_DISPLAY = 10

# Set to FALSE to disable SSL verification to use tools like charles, mitmproxy, fiddler, or similiar tools to debug the data sent on the wire.
# Can also use command-line argument --disable-ssl-verify
# This uses code taken from:
# - https://stackoverflow.com/questions/15445981/how-do-i-disable-the-security-certificate-check-in-python-requests
# - https://stackoverflow.com/questions/16337511/log-all-requests-from-the-python-requests-module
ENABLE_SSL_VERIFICATION = True

# Set the log level for HTTP commands
# HTTP_LOG_LEVEL = logging.CRITICAL
HTTP_LOG_LEVEL = logging.ERROR
# HTTP_LOG_LEVEL = logging.WARNING
# HTTP_LOG_LEVEL = logging.INFO
# HTTP_LOG_LEVEL = logging.DEBUG

# Set the log level for other commands
# OTHER_LOG_LEVEL = logging.CRITICAL
OTHER_LOG_LEVEL = logging.ERROR
# OTHER_LOG_LEVEL = logging.WARNING
# OTHER_LOG_LEVEL = logging.INFO
# OTHER_LOG_LEVEL = logging.DEBUG

# Set whether to show debug info for HTTPConnection
HTTPCONNECTION_DEBUG_INFO = False

# Set where you'd like the COOKIES to be stored. Can also use command-line argument --cookie-dir
COOKIE_DIR = ""  # location to store session information

# Other configurable variables
APPLE_USERNAME = ""
APPLE_PASSWORD = ""
CHINA = False


def parse_args() -> argparse.Namespace:
    """Parse command line arguments"""
    global ENABLE_SSL_VERIFICATION, COOKIE_DIR, APPLE_PASSWORD, APPLE_USERNAME, CHINA
    parser = argparse.ArgumentParser(description="End to End Test of Services")

    parser.add_argument(
        "--username",
        action="store",
        dest="username",
        default="",
        help="Apple ID to Use",
    )
    parser.add_argument(
        "--password",
        action="store",
        dest="password",
        default="",
        help=(
            "Apple ID Password to Use; if unspecified, password will be "
            "fetched from the system keyring."
        ),
    )
    parser.add_argument(
        "--cookie-dir",
        action="store",
        dest="cookie_directory",
        default="",
        help="Directory to store session information and cookies",
    )
    parser.add_argument(
        "--china-mainland",
        action="store_true",
        dest="china_mainland",
        default=False,
        help="If the country/region setting of the Apple ID is China mainland",
    )
    parser.add_argument(
        "--disable-ssl-verify",
        action="store_true",
        dest="disable_ssl",
        default=False,
        help="Disable SSL verification",
    )

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    args = parser.parse_args()

    if not args.username or not args.password:
        parser.error("Both --username and --password are required")
    else:
        APPLE_USERNAME = args.username
        APPLE_PASSWORD = args.password

    if args.cookie_directory:
        COOKIE_DIR = args.cookie_directory

    if args.china_mainland:
        CHINA = args.china_mainland

    if args.disable_ssl or not ENABLE_SSL_VERIFICATION:
        ENABLE_SSL_VERIFICATION = False
        print("=" * 80)
        print("⚠️  SECURITY WARNING: SSL VERIFICATION DISABLED ⚠️")
        print("This is insecure and should ONLY be used for debugging!")
        print("Your credentials and data may be exposed to attackers.")
        print("=" * 80)
        print()

    return args


@contextlib.contextmanager
def configurable_ssl_verification(verify_ssl=True):
    opened_adapters = set()

    def merge_environment_settings_with_config(
        self, url, proxies, stream, verify, cert
    ):
        # Add opened adapters to a set so they can be closed later
        opened_adapters.add(self.get_adapter(url))

        settings = old_merge_environment_settings(
            self, url, proxies, stream, verify, cert
        )

        if not verify_ssl:
            settings["verify"] = False
            # You can also uncomment and use proxies here if needed,
            # proxies = {
            #     "http": "http://127.0.0.1:8888",
            #     "https": "http://127.0.0.1:8888"
            # }
            # settings["proxies"] = proxies

        return settings

    # Temporarily override merge_environment_settings
    requests.Session.merge_environment_settings = merge_environment_settings_with_config

    try:
        # Only catch InsecureRequestWarning if we are disabling SSL verification
        if not verify_ssl:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", InsecureRequestWarning)
                yield
        else:
            yield
    finally:
        # Restore the original merge_environment_settings
        requests.Session.merge_environment_settings = old_merge_environment_settings

        # Close all opened adapters
        for adapter in opened_adapters:
            try:
                adapter.close()
            except Exception:
                pass  # Ignore errors during adapter closing


def httpclient_logging_patch(level=HTTP_LOG_LEVEL):
    """Enable HTTPConnection debug logging to the logging framework"""

    def httpclient_log(*args):
        httpclient_logger.log(level, " ".join(args))

    # mask the print() built-in in the http.client module to use
    # logging instead
    http.client.print = httpclient_log
    # enable debugging
    if HTTPCONNECTION_DEBUG_INFO:
        http.client.HTTPConnection.debuglevel = 1
    else:
        http.client.HTTPConnection.debuglevel = 0


def handle_2fa(api: PyiCloudService) -> None:
    """Handle two-factor authentication"""
    security_key_names: Optional[List[str]] = api.security_key_names

    with configurable_ssl_verification(ENABLE_SSL_VERIFICATION):
        if security_key_names:
            print(
                f"Security key confirmation is required. "
                f"Please plug in one of the following keys: {', '.join(security_key_names)}"
            )

            fido2_devices: List[CtapHidDevice] = api.fido2_devices

            print("Available FIDO2 devices:")

            for idx, dev in enumerate(fido2_devices, start=1):
                print(f"{idx}: {dev}")

            choice = click.prompt(
                "Select a FIDO2 device by number",
                type=click.IntRange(1, len(fido2_devices)),
                default=1,
            )
            selected_device: CtapHidDevice = fido2_devices[choice - 1]

            print("Please confirm the action using the security key")

            api.confirm_security_key(selected_device)

        else:
            print("Two-factor authentication required.")
            code: str = input(
                "Enter the code you received of one of your approved devices: "
            )
            result: bool = api.validate_2fa_code(code)
            print(f"Code validation result: {result}")

            if not result:
                print("Failed to verify security code")
                sys.exit(1)

        if not api.is_trusted_session:
            print("Session is not trusted. Requesting trust...")
            result = api.trust_session()
            print(f"Session trust result: {result}")

            if not result:
                print(
                    "Failed to request trust. You will likely be prompted for confirmation again in the coming weeks"
                )


def handle_2sa(api: PyiCloudService) -> None:
    """Handle two-step authentication"""
    print("Two-step authentication required. Your trusted devices are:")

    with configurable_ssl_verification(ENABLE_SSL_VERIFICATION):
        trusted_devices: List[dict[str, Any]] = api.trusted_devices
        for i, device in enumerate(trusted_devices):
            print(
                "  %s: %s"
                % (i, device.get("deviceName", "SMS to %s" % device.get("phoneNumber")))
            )

        device_index: int = click.prompt(
            "Which device would you like to use?", default=0
        )
        device: dict[str, Any] = trusted_devices[device_index]
        if not api.send_verification_code(device):
            print("Failed to send verification code")
            sys.exit(1)

        code = click.prompt("Please enter validation code")
        if not api.validate_verification_code(device, code):
            print("Failed to verify verification code")
            sys.exit(1)


def get_api() -> PyiCloudService:
    parse_args()

    with configurable_ssl_verification(ENABLE_SSL_VERIFICATION):
        api = PyiCloudService(
            apple_id=APPLE_USERNAME,
            password=APPLE_PASSWORD,
            china_mainland=CHINA,
            cookie_directory=COOKIE_DIR,
        )

    if api.requires_2fa:
        handle_2fa(api)

    elif api.requires_2sa:
        handle_2sa(api)

    return api


def display_devices(api: PyiCloudService) -> None:
    """Display device info"""
    with configurable_ssl_verification(ENABLE_SSL_VERIFICATION):
        print(f"List of devices ({len(api.devices)}):")
        for idx, device in enumerate(api.devices):
            print(f"\t{idx}: {device}")
            if idx >= MAX_DISPLAY - 1:
                break
        print(END_LIST)

        print("First device:")
        print(f"\t Name: {api.iphone}")
        print(f"\t Location: {json.dumps(api.iphone.location, indent=4)}\n")


def display_calendars(api: PyiCloudService) -> None:
    """Display calendar info"""
    with configurable_ssl_verification(ENABLE_SSL_VERIFICATION):
        calendar_service: CalendarService = api.calendar
        calendars: list[CalendarObject] = calendar_service.get_calendars(as_objs=True)
        print(f"List of calendars ({len(calendars)}):")
        for idx, calendar in enumerate(calendars):
            print(f"\t{idx}: {calendar.title}")
            if idx >= MAX_DISPLAY - 1:
                break
        print(END_LIST)

        if calendars:
            # Get recent events from API
            try:
                recent_events = calendar_service.get_events(
                    from_dt=datetime.now() - timedelta(days=7),
                    to_dt=datetime.now() + timedelta(days=7),
                    as_objs=True,
                )
                print(f"Recent events (±7 days): {len(recent_events)} events")
                for idx, event in enumerate(recent_events):
                    if hasattr(event, "title") and hasattr(event, "start_date"):
                        print(f"\t{idx}: {event.title} at {event.start_date}")
                        if idx >= MAX_DISPLAY - 1:
                            break
                print(END_LIST)
            except Exception as e:
                print(f"Could not retrieve events: {e}\n")


def display_contacts(api: PyiCloudService) -> None:
    """Display contacts info"""
    with configurable_ssl_verification(ENABLE_SSL_VERIFICATION):
        contacts = api.contacts.all
        if contacts:
            print(f"List of contacts ({len(contacts)}):")
            for idx, contact in enumerate(contacts):
                print(
                    f"\t{idx}: {contact.get('firstName') or contact.get('lastName') or contact.get('companyName')}"
                )
                if idx >= MAX_DISPLAY - 1:
                    break
            print(END_LIST)
        else:
            print("No contacts found\n")


def display_drive(api: PyiCloudService) -> None:
    """Display drive info"""
    with configurable_ssl_verification(ENABLE_SSL_VERIFICATION):
        drive_files: list[str] = api.drive.dir()
        print(f"List of files in iCloud Drive root ({len(drive_files)}):")
        for idx, filename in enumerate(drive_files):
            print(f"\t{idx}: {filename} ({api.drive[filename].type})")
            if idx >= MAX_DISPLAY - 1:
                break
        print(END_LIST)


def display_files(api: PyiCloudService) -> None:
    """Display files info"""
    with configurable_ssl_verification(ENABLE_SSL_VERIFICATION):
        try:
            files: list[str] = api.files.dir()
            print(f"List of files in iCloud files root ({len(files)}):")
            for idx, filename in enumerate(files):
                print(f"\t{idx}: {filename} ({api.files[filename].type})")
                if idx >= MAX_DISPLAY - 1:
                    break
            print(END_LIST)
        except PyiCloudServiceUnavailable as error:
            print(f"Files service not available: {error}\n")


def display_photos(api: PyiCloudService) -> None:
    """Display photo info"""
    with configurable_ssl_verification(ENABLE_SSL_VERIFICATION):
        print(f"List of photo albums ({len(api.photos.albums)}):")
        for idx, album in enumerate(api.photos.albums):
            print(f"\t{idx}: {album}")
            if idx >= MAX_DISPLAY - 1:
                break
        print(END_LIST)

        print(f"List of ALL PHOTOS ({len(api.photos.all)}):")
        for idx, photo in enumerate(api.photos.all):
            print(f"\t{idx}: {photo.filename} ({photo.item_type})")
            if idx >= MAX_DISPLAY - 1:
                break
        print(END_LIST)


def display_videos(api: PyiCloudService) -> None:
    """Display video info"""
    with configurable_ssl_verification(ENABLE_SSL_VERIFICATION):
        print(f"List of Videos ({len(api.photos.albums['Videos'])}):")
        for idx, photo in enumerate(api.photos.albums["Videos"]):
            print(f"\t{idx}: {photo.filename} ({photo.item_type})")
            if idx >= MAX_DISPLAY - 1:
                break
        print(END_LIST)


def display_shared_photos(api: PyiCloudService) -> None:
    """Display shared photo info"""
    with configurable_ssl_verification(ENABLE_SSL_VERIFICATION):
        album = None
        print(f"List of Shared Albums ({len(api.photos.shared_streams)}):")
        for idx, album in enumerate(api.photos.shared_streams):
            print(f"\t{idx}: {album}")
            if idx >= MAX_DISPLAY - 1:
                break
        print(END_LIST)

        if album and api.photos.shared_streams:
            print(
                f"List of Shared Photos [{album}] ({len(api.photos.shared_streams[album])}):"
            )
            for idx, photo in enumerate(api.photos.shared_streams[album]):
                print(f"\t{idx}: {photo.filename} ({photo.item_type})")

                if idx >= MAX_DISPLAY - 1:
                    break
            print(END_LIST)


def display_account(api: PyiCloudService) -> None:
    """Display account info"""
    with configurable_ssl_verification(ENABLE_SSL_VERIFICATION):
        print(f"Account name: {api.account_name}")
        print(f"Account plan: {json.dumps(api.account.summary_plan, indent=4)}")
        print(f"List of Family Member ({len(api.account.family)}):")
        for idx, member in enumerate(api.account.family):
            print(f"\t{idx}: {member}")
            try:
                photo: Response = member.get_photo()
                print(f"\t\tPhoto: {photo}")
                print(f"\t\tPhoto type: {photo.headers['Content-Type']}")
                print(f"\t\tPhoto size: {photo.headers['Content-Length']}")
            except Exception as e:
                print(f"\t\tPhoto: Error retrieving user photo: {e}")
            if idx >= MAX_DISPLAY - 1:
                break
        print(END_LIST)


def display_hidemyemail(api: PyiCloudService) -> None:
    """Display Hide My Email info"""
    with configurable_ssl_verification(ENABLE_SSL_VERIFICATION):
        print(f"List of Hide My Email ({len(api.hidemyemail)}):")
        for idx, email in enumerate(api.hidemyemail):
            print(
                f"\t{idx}: {email['hme']} ({email['domain']}) Active = {email['isActive']}"
            )
            if idx >= MAX_DISPLAY - 1:
                break
        print(END_LIST)


def main() -> None:
    global httpclient_logger, old_merge_environment_settings

    # Store the original merge_environment_settings
    old_merge_environment_settings = requests.Session.merge_environment_settings

    # Enable general debug logging
    logging.basicConfig(level=OTHER_LOG_LEVEL)

    # Enable httpclient logging
    httpclient_logger = logging.getLogger("http.client")
    httpclient_logging_patch()

    """main function"""
    api: PyiCloudService = get_api()

    display_account(api)
    display_devices(api)
    display_hidemyemail(api)
    try:
        display_calendars(api)
    except PyiCloudServiceUnavailable as error:
        print(f"Calendar service not available: {error}\n")
    display_files(api)
    display_contacts(api)
    display_drive(api)
    display_photos(api)
    display_videos(api)
    display_shared_photos(api)


if __name__ == "__main__":
    main()



"""Find My iPhone service tests."""

# pylint: disable=protected-access

from unittest.mock import patch

from pyicloud.base import PyiCloudService
from pyicloud.services.findmyiphone import AppleDevice, FindMyiPhoneServiceManager


def test_devices(pyicloud_service_working: PyiCloudService) -> None:
    """Tests devices."""

    assert pyicloud_service_working.devices

    for device in pyicloud_service_working.devices:
        assert device["canWipeAfterLock"] is not None
        assert device["baUUID"] is not None
        assert device["wipeInProgress"] is not None
        assert device["lostModeEnabled"] is not None
        assert device["activationLocked"] is not None
        assert device["passcodeLength"] is not None
        assert device["deviceStatus"] is not None
        assert device["features"] is not None
        assert device["lowPowerMode"] is not None
        assert device["rawDeviceModel"] is not None
        assert device["id"] is not None
        assert device["isLocating"] is not None
        assert device["modelDisplayName"] is not None
        assert device["lostTimestamp"] is not None
        assert device["batteryLevel"] is not None
        assert device["locationEnabled"] is not None
        assert device["locFoundEnabled"] is not None
        assert device["fmlyShare"] is not None
        assert device["lostModeCapable"] is not None
        assert device["wipedTimestamp"] is None
        assert device["deviceDisplayName"] is not None
        assert device["audioChannels"] is not None
        assert device["locationCapable"] is not None
        assert device["batteryStatus"] is not None
        assert device["trackingInfo"] is None
        assert device["name"] is not None
        assert device["isMac"] is not None
        assert device["thisDevice"] is not None
        assert device["deviceClass"] is not None
        assert device["deviceModel"] is not None
        assert device["maxMsgChar"] is not None
        assert device["darkWake"] is not None
        assert device["remoteWipe"] is None

        assert device.data["canWipeAfterLock"] is not None
        assert device.data["baUUID"] is not None
        assert device.data["wipeInProgress"] is not None
        assert device.data["lostModeEnabled"] is not None
        assert device.data["activationLocked"] is not None
        assert device.data["passcodeLength"] is not None
        assert device.data["deviceStatus"] is not None
        assert device.data["features"] is not None
        assert device.data["lowPowerMode"] is not None
        assert device.data["rawDeviceModel"] is not None
        assert device.data["id"] is not None
        assert device.data["isLocating"] is not None
        assert device.data["modelDisplayName"] is not None
        assert device.data["lostTimestamp"] is not None
        assert device.data["batteryLevel"] is not None
        assert device.data["locationEnabled"] is not None
        assert device.data["locFoundEnabled"] is not None
        assert device.data["fmlyShare"] is not None
        assert device.data["lostModeCapable"] is not None
        assert device.data["wipedTimestamp"] is None
        assert device.data["deviceDisplayName"] is not None
        assert device.data["audioChannels"] is not None
        assert device.data["locationCapable"] is not None
        assert device.data["batteryStatus"] is not None
        assert device.data["trackingInfo"] is None
        assert device.data["name"] is not None
        assert device.data["isMac"] is not None
        assert device.data["thisDevice"] is not None
        assert device.data["deviceClass"] is not None
        assert device.data["deviceModel"] is not None
        assert device.data["maxMsgChar"] is not None
        assert device.data["darkWake"] is not None
        assert device.data["remoteWipe"] is None


def test_apple_device_properties(pyicloud_service_working: PyiCloudService) -> None:
    """Tests AppleDevice properties and methods."""
    device: AppleDevice = pyicloud_service_working.devices[0]

    # Test session property
    assert device.session is not None

    # Test location property
    location = device.location
    assert location is not None
    assert "latitude" in location
    assert "longitude" in location

    # Test status method
    status = device.status()
    assert "batteryLevel" in status
    assert "deviceDisplayName" in status
    assert "deviceStatus" in status
    assert "name" in status

    # Test status with additional fields
    additional_status = device.status(additional=["isMac", "deviceClass"])
    assert "isMac" in additional_status
    assert "deviceClass" in additional_status

    # Test data property
    assert device.data is not None
    assert "id" in device.data

    # Test __getitem__ method
    assert device["id"] == device.data["id"]

    # Test __getattr__ method
    assert device.deviceDisplayName == device.data["deviceDisplayName"]

    # Test __str__ method
    assert str(device) == f"{device['deviceDisplayName']}: {device['name']}"

    # Test __repr__ method
    assert repr(device) == f"<AppleDevice({device})>"


def test_apple_device_actions(pyicloud_service_working: PyiCloudService) -> None:
    """Tests AppleDevice actions like play_sound, display_message, and lost_device."""
    device: AppleDevice = pyicloud_service_working.devices[0]

    # Mock session.post to avoid actual API calls
    with patch.object(device.session, "post") as mock_post:
        # Test play_sound
        device.play_sound(subject="Test Alert")
        mock_post.assert_called_with(
            device._sound_url,
            params=device._params,
            json={
                "device": device.data["id"],
                "subject": "Test Alert",
                "clientContext": {"fmly": True},
            },
        )

        # Test display_message
        device.display_message(subject="Test Message", message="Hello", sounds=True)
        mock_post.assert_called_with(
            device._message_url,
            params=device._params,
            json={
                "device": device.data["id"],
                "subject": "Test Message",
                "sound": True,
                "userText": True,
                "text": "Hello",
            },
        )

        # Test lost_device
        device.lost_device(
            number="1234567890", text="Lost device message", newpasscode="1234"
        )
        mock_post.assert_called_with(
            device._lost_url,
            params=device._params,
            json={
                "text": "Lost device message",
                "userText": True,
                "ownerNbr": "1234567890",
                "lostModeEnabled": True,
                "trackingEnabled": True,
                "device": device.data["id"],
                "passcode": "1234",
            },
        )


def test_findmyiphone_service_manager(
    pyicloud_service_working: PyiCloudService,
) -> None:
    """Tests FindMyiPhoneServiceManager methods."""
    manager: FindMyiPhoneServiceManager = pyicloud_service_working.devices

    # Test refresh_client
    manager.refresh_client()
    assert len(manager) > 0

    # Test __getitem__
    device = manager[0]
    assert isinstance(device, AppleDevice)

    # Test __len__
    assert len(manager) == len(manager)

    # Test __iter__
    devices = list(iter(manager))
    assert len(devices) == len(manager)

    # Test __str__ and __repr__
    assert str(manager) == repr(manager)



"""Pyicloud Session handling"""

import http.cookiejar
import logging
import os
import os.path as path
from json import JSONDecodeError, dump, load
from re import match
from typing import Any, NoReturn, Optional, Union, cast

import requests
import requests.cookies
from requests.models import Response

from pyicloud.const import (
    CONTENT_TYPE,
    CONTENT_TYPE_JSON,
    CONTENT_TYPE_TEXT_JSON,
    ERROR_ACCESS_DENIED,
    ERROR_AUTHENTICATION_FAILED,
    ERROR_ZONE_NOT_FOUND,
    HEADER_DATA,
)
from pyicloud.exceptions import (
    PyiCloud2FARequiredException,
    PyiCloud2SARequiredException,
    PyiCloudAPIResponseException,
    PyiCloudServiceNotActivatedException,
)

KEY_RETRIED = "retried"


class PyiCloudCookieJar(
    requests.cookies.RequestsCookieJar, http.cookiejar.LWPCookieJar
):
    """Mix the Requests CookieJar with the LWPCookieJar to allow persistance"""


class PyiCloudSession(requests.Session):
    """iCloud session."""

    def __init__(
        self,
        service,
        client_id: str,
        cookie_directory: str,
        verify: bool = False,
        headers: Optional[dict[str, str]] = None,
    ) -> None:
        super().__init__()

        self._service = service
        self.verify = verify
        self._cookie_directory: str = cookie_directory
        self.cookies = PyiCloudCookieJar(self.cookiejar_path)
        self._data: dict[str, Any] = {}

        self._logger: logging.Logger = logging.getLogger(__name__)

        if headers:
            self.headers.update(headers)

        self._load_session_data()

        if not self._data.get("client_id"):
            self._data.update({"client_id": client_id})

    @property
    def data(self) -> dict[str, Any]:
        """Gets the session data"""
        return self._data

    @property
    def logger(self) -> logging.Logger:
        """Gets the request logger"""
        if (
            self.service.password_filter is not None
            and self.service.password_filter not in self._logger.filters
        ):
            self._logger.addFilter(self.service.password_filter)
        return self._logger

    def _load_session_data(self) -> None:
        """Load session_data from file."""
        if os.path.exists(self.cookiejar_path):
            cast(PyiCloudCookieJar, self.cookies).load(
                ignore_discard=True, ignore_expires=True
            )

        self._logger.debug("Using session file %s", self.session_path)
        self._data: dict[str, Any] = {}
        try:
            with open(self.session_path, encoding="utf-8") as session_f:
                self._data = load(session_f)
        except (
            JSONDecodeError,
            OSError,
        ):
            self._logger.info("Session file does not exist")

    def _save_session_data(self) -> None:
        """Save session_data to file."""
        with open(self.session_path, "w", encoding="utf-8") as outfile:
            dump(self._data, outfile)
            self.logger.debug("Saved session data to file: %s", self.session_path)

        cast(PyiCloudCookieJar, self.cookies).save(
            ignore_discard=True, ignore_expires=True
        )
        self.logger.debug("Saved cookies data to file: %s", self.cookiejar_path)

    def _update_session_data(self, response: Response) -> None:
        """Update session_data with new data."""
        for header, value in HEADER_DATA.items():
            if response.headers.get(header):
                session_arg: str = value
                self._data.update({session_arg: response.headers.get(header)})

    def _is_json_response(self, response: Response) -> bool:
        content_type: str = response.headers.get(CONTENT_TYPE, "")
        json_mimetypes: list[str] = [
            CONTENT_TYPE_JSON,
            CONTENT_TYPE_TEXT_JSON,
        ]
        return content_type.split(";")[0] in json_mimetypes

    def _reauthenticate_find_my_iphone(self, response: Response) -> None:
        self.logger.debug("Re-authenticating Find My iPhone service")
        try:
            service: Optional[str] = None if response.status_code == 450 else "find"
            self.service.authenticate(True, service)
        except PyiCloudAPIResponseException:
            self.logger.debug("Re-authentication failed")

    def request(
        self,
        method,
        url,
        params=None,
        data=None,
        headers=None,
        cookies=None,
        files=None,
        auth=None,
        timeout=None,
        allow_redirects=True,
        proxies=None,
        hooks=None,
        stream=None,
        verify=None,
        cert=None,
        json=None,
    ) -> Response:
        return self._request(
            method,
            url,
            params=params,
            data=data,
            headers=headers,
            cookies=cookies,
            files=files,
            auth=auth,
            timeout=timeout,
            allow_redirects=allow_redirects,
            proxies=proxies,
            hooks=hooks,
            stream=stream,
            verify=verify,
            cert=cert,
            json=json,
        )

    def _request(
        self,
        method,
        url,
        *,
        data=None,
        has_retried: bool = False,
        **kwargs,
    ) -> Response:
        """Request method."""
        self.logger.debug(
            "%s %s %s",
            method,
            url,
            data or "",
        )

        try:
            response: Response = super().request(
                method=method,
                url=url,
                data=data,
                **kwargs,
            )

            self._update_session_data(response)
            self._save_session_data()

            if not response.ok and (
                self._is_json_response(response)
                or response.status_code in [421, 450, 500]
            ):
                return self._handle_request_error(
                    response=response,
                    method=method,
                    url=url,
                    data=data,
                    has_retried=has_retried,
                    **kwargs,
                )

            response.raise_for_status()

            if not self._is_json_response(response):
                return response

            self._decode_json_response(response)

            return response
        except requests.HTTPError as err:
            raise PyiCloudAPIResponseException(
                reason=err.response.text,
                code=err.response.status_code,
            ) from err
        except requests.exceptions.RequestException as err:
            raise PyiCloudAPIResponseException("Request failed to iCloud") from err

    def _handle_request_error(
        self,
        response: Response,
        method,
        url,
        *,
        data=None,
        has_retried: bool = False,
        **kwargs,
    ) -> Response:
        """Handle request error."""

        if (
            response.status_code == 409
            and self._is_json_response(response)
            and (response.json().get("authType") == "hsa2")
        ):
            raise PyiCloud2FARequiredException(
                apple_id=self.service.account_name,
                response=response,
            )

        try:
            fmip_url: str = self.service.get_webservice_url("findme")
            if (
                not has_retried
                and response.status_code in [421, 450, 500]
                and fmip_url in url
            ):
                self._reauthenticate_find_my_iphone(response)
                return self._request(
                    method=method,
                    url=url,
                    data=data,
                    has_retried=True,
                    **kwargs,
                )
        except PyiCloudServiceNotActivatedException:
            pass

        if not has_retried and response.status_code in [421, 450, 500]:
            return self._request(
                method=method,
                url=url,
                data=data,
                has_retried=True,
                **kwargs,
            )

        self._raise_error(response.status_code, response.reason)

    def _decode_json_response(self, response: Response) -> None:
        """Decode JSON response."""
        if len(response.content) == 0:
            return

        try:
            data: Union[list[dict[str, Any]], dict[str, Any]] = response.json()
            if isinstance(data, dict):
                reason: Optional[str] = data.get("errorMessage")
                reason = reason or data.get("reason")
                reason = reason or data.get("errorReason")
                reason = reason or data.get("error")
                if reason and not isinstance(reason, str):
                    reason = "Unknown reason"

                if reason:
                    code: Optional[Union[int, str]] = data.get("errorCode")
                    code = code or data.get("serverErrorCode")
                    self._raise_error(code, reason)

        except JSONDecodeError:
            self.logger.warning(
                "Failed to parse response with JSON mimetype: %s", response.text
            )

    def _raise_error(self, code: Optional[Union[int, str]], reason: str) -> NoReturn:
        if (
            self.service.requires_2sa
            and reason == "Missing X-APPLE-WEBAUTH-TOKEN cookie"
        ):
            raise PyiCloud2SARequiredException(self.service.account_name)
        if code in (ERROR_ZONE_NOT_FOUND, ERROR_AUTHENTICATION_FAILED):
            reason = (
                "Please log into https://icloud.com/ to manually "
                "finish setting up your iCloud service"
            )
            raise PyiCloudServiceNotActivatedException(reason, code)
        if code == ERROR_ACCESS_DENIED:
            reason = (
                reason + ".  Please wait a few minutes then try again."
                "The remote servers might be trying to throttle requests."
            )
        if code in [421, 450, 500]:
            reason = "Authentication required for Account."

        raise PyiCloudAPIResponseException(reason, code)

    @property
    def service(self):
        """Gets the service."""
        return self._service

    @property
    def cookiejar_path(self) -> str:
        """Get path for cookiejar file."""
        return path.join(
            self._cookie_directory,
            "".join([c for c in self.service.account_name if match(r"\w", c)])
            + ".cookiejar",
        )

    @property
    def session_path(self) -> str:
        """Get path for session data file."""
        return path.join(
            self._cookie_directory,
            "".join([c for c in self.service.account_name if match(r"\w", c)])
            + ".session",
        )



"""
Test the PyiCloudService and PyiCloudSession classes."""

# pylint: disable=protected-access

from unittest.mock import MagicMock, mock_open, patch

import pytest
from requests import Response

from pyicloud.base import PyiCloudService, PyiCloudSession, b64_encode
from pyicloud.exceptions import (
    PyiCloud2SARequiredException,
    PyiCloudAPIResponseException,
    PyiCloudFailedLoginException,
    PyiCloudServiceNotActivatedException,
)


def test_authenticate_with_force_refresh(pyicloud_service: PyiCloudService) -> None:
    """Test the authenticate method with force_refresh=True."""
    with (
        patch("pyicloud.base.PyiCloudSession.post") as mock_post_response,
        patch("pyicloud.base.PyiCloudService._validate_token") as validate_token,
    ):
        pyicloud_service.session._data = {"session_token": "valid_token"}
        mock_post_response.json.return_value = {
            "apps": {"test_service": {"canLaunchWithOneFactor": True}},
            "status": "success",
        }
        pyicloud_service.data = {
            "apps": {"test_service": {"canLaunchWithOneFactor": True}}
        }
        validate_token = MagicMock(
            return_value={
                "status": "success",
                "dsInfo": {"hsaVersion": 1},
                "webservices": "TestWebservices",
            }
        )
        pyicloud_service._validate_token = validate_token
        pyicloud_service.authenticate(force_refresh=True, service="test_service")
        mock_post_response.assert_called_once()
        validate_token.assert_called_once()


def test_authenticate_with_missing_token(pyicloud_service: PyiCloudService) -> None:
    """Test the authenticate method with missing session_token."""
    with (
        patch("pyicloud.base.PyiCloudSession.post") as mock_post_response,
        patch.object(
            pyicloud_service,
            "_authenticate_with_token",
            side_effect=[PyiCloudFailedLoginException, None],
        ) as mock_authenticate_with_token,
    ):
        mock_post_response.return_value.json.side_effect = [
            {
                "salt": "U29tZVNhbHQ=",
                "b": "U29tZUJ5dGVz",
                "c": "TestC",
                "iteration": 1000,
                "dsInfo": {"hsaVersion": 1},
                "hsaChallengeRequired": False,
                "webservices": "TestWebservices",
            },
            None,
        ]
        pyicloud_service.session.post = mock_post_response
        pyicloud_service.session._data = {}
        pyicloud_service.params = {}
        pyicloud_service.authenticate()
        assert mock_post_response.call_count == 2
        assert mock_authenticate_with_token.call_count == 2


def test_validate_2fa_code(pyicloud_service: PyiCloudService) -> None:
    """Test the validate_2fa_code method with a valid code."""

    pyicloud_service.data = {"dsInfo": {"hsaVersion": 1}, "hsaChallengeRequired": False}

    with patch("pyicloud.base.PyiCloudSession") as mock_session:
        pyicloud_service.session = mock_session
        mock_session.data = {
            "scnt": "test_scnt",
            "session_id": "test_session_id",
            "session_token": "test_session_token",
        }

        mock_post_response = MagicMock()
        mock_post_response.status_code = 200
        mock_post_response.json.return_value = {"success": True}
        mock_session.post.return_value = mock_post_response

        assert pyicloud_service.validate_2fa_code("123456")


def test_validate_2fa_code_failure(pyicloud_service: PyiCloudService) -> None:
    """Test the validate_2fa_code method with an invalid code."""
    exception = PyiCloudAPIResponseException("Invalid code")
    exception.code = -21669
    with patch("pyicloud.base.PyiCloudSession") as mock_session:
        mock_session.post.side_effect = exception
        pyicloud_service.session = mock_session
        assert not pyicloud_service.validate_2fa_code("000000")


@patch("pyicloud.base.CtapHidDevice.list_devices", return_value=[MagicMock()])
@patch("pyicloud.base.Fido2Client")
def test_confirm_security_key_success(
    mock_fido2_client_cls, mock_list_devices, pyicloud_service: PyiCloudService
) -> None:
    """Test that the FIDO2 WebAuthn flow works"""
    rp_id = "example.com"
    challenge = "ZmFrZV9jaGFsbGVuZ2U"

    # Arrange
    pyicloud_service._submit_webauthn_assertion_response = MagicMock()
    pyicloud_service.trust_session = MagicMock()

    # Simulated WebAuthn options returned from backend
    pyicloud_service._get_webauthn_options = MagicMock(
        return_value={
            "fsaChallenge": {
                "challenge": challenge,  # base64url(fake_challenge)
                "keyHandles": ["a2V5MQ", "a2V5Mg"],  # base64url(fake_key_ids)
                "rpId": rp_id,
            }
        }
    )

    # Simulated FIDO2 response
    mock_response = MagicMock()
    mock_response.response = MagicMock()
    mock_response.response.client_data = b"client_data"
    mock_response.response.signature = b"signature"
    mock_response.response.authenticator_data = b"auth_data"
    mock_response.response.user_handle = b"user_handle"
    mock_response.raw_id = b"cred_id"

    mock_fido2_client = MagicMock()
    mock_fido2_client.get_assertion.return_value.get_response.return_value = (
        mock_response
    )
    mock_fido2_client_cls.return_value = mock_fido2_client

    # Act
    pyicloud_service.confirm_security_key()

    # Assert
    mock_list_devices.assert_called_once()
    mock_fido2_client.get_assertion.assert_called_once()

    # Check if data was submitted correctly
    pyicloud_service._submit_webauthn_assertion_response.assert_called_once_with(
        {
            "challenge": challenge,
            "rpId": rp_id,
            "clientData": b64_encode(mock_response.response.client_data),
            "signatureData": b64_encode(mock_response.response.signature),
            "authenticatorData": b64_encode(mock_response.response.authenticator_data),
            "userHandle": b64_encode(mock_response.response.user_handle),
            "credentialID": b64_encode(mock_response.raw_id),
        }
    )

    pyicloud_service.trust_session.assert_called_once()


def test_get_webservice_url_success(pyicloud_service: PyiCloudService) -> None:
    """Test the get_webservice_url method with a valid key."""
    pyicloud_service._webservices = {"test_key": {"url": "https://example.com"}}
    url: str = pyicloud_service.get_webservice_url("test_key")
    assert url == "https://example.com"


def test_get_webservice_url_failure(pyicloud_service: PyiCloudService) -> None:
    """Test the get_webservice_url method with an invalid key."""
    pyicloud_service._webservices = {}
    with pytest.raises(PyiCloudServiceNotActivatedException):
        pyicloud_service.get_webservice_url("invalid_key")


def test_trust_session_success(pyicloud_service: PyiCloudService) -> None:
    """Test the trust_session method with a successful response."""

    with patch("pyicloud.base.PyiCloudSession") as mock_session:
        mock_session.data = {
            "scnt": "test_scnt",
            "session_id": "test_session_id",
            "session_token": "test_session_token",
        }
        pyicloud_service.session = mock_session
        assert pyicloud_service.trust_session()


def test_trust_session_failure(pyicloud_service: PyiCloudService) -> None:
    """Test the trust_session method with a failed response."""
    with patch("pyicloud.base.PyiCloudSession") as mock_session:
        pyicloud_service.session = mock_session
        mock_session.get.side_effect = PyiCloudAPIResponseException("Trust failed")
        assert not pyicloud_service.trust_session()


def test_cookiejar_path_property(pyicloud_session: PyiCloudSession) -> None:
    """Test the cookiejar_path property."""
    path: str = pyicloud_session.cookiejar_path
    assert isinstance(path, str)


def test_session_path_property(pyicloud_session: PyiCloudSession) -> None:
    """Test the session_path property."""
    path: str = pyicloud_session.session_path
    assert isinstance(path, str)


def test_requires_2sa_property(pyicloud_service: PyiCloudService) -> None:
    """Test the requires_2sa property."""
    pyicloud_service.data = {"dsInfo": {"hsaVersion": 2}}
    assert pyicloud_service.requires_2sa


def test_requires_2fa_property(pyicloud_service: PyiCloudService) -> None:
    """Test the requires_2fa property."""
    pyicloud_service.data = {"dsInfo": {"hsaVersion": 2}, "hsaChallengeRequired": False}
    assert pyicloud_service.requires_2fa


def test_is_trusted_session_property(pyicloud_service: PyiCloudService) -> None:
    """Test the is_trusted_session property."""
    pyicloud_service.data = {"dsInfo": {"hsaVersion": 2}}
    assert not pyicloud_service.is_trusted_session


def test_request_success(pyicloud_service_working: PyiCloudService) -> None:
    """Test the request method with a successful response."""
    with (
        patch("requests.Session.request") as mock_request,
        patch("builtins.open", new_callable=mock_open),
        patch("http.cookiejar.LWPCookieJar.save") as mock_save,
    ):
        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.json.return_value = {"success": True}
        mock_response.headers.get.return_value = "application/json"
        mock_request.return_value = mock_response
        pyicloud_session = PyiCloudSession(
            pyicloud_service_working, "", cookie_directory=""
        )

        response: Response = pyicloud_session.request(
            "POST", "https://example.com", data={"key": "value"}
        )
        assert response.json() == {"success": True}
        assert response.headers.get("Content-Type") == "application/json"
        mock_request.assert_called_once_with(
            method="POST",
            url="https://example.com",
            data={"key": "value"},
            params=None,
            headers=None,
            cookies=None,
            files=None,
            auth=None,
            timeout=None,
            allow_redirects=True,
            proxies=None,
            hooks=None,
            stream=None,
            verify=None,
            cert=None,
            json=None,
        )
        mock_save.assert_called_once()


def test_request_failure(pyicloud_service_working: PyiCloudService) -> None:
    """Test the request method with a failure response."""

    with (
        patch("requests.Session.request") as mock_request,
        patch("builtins.open", new_callable=mock_open) as open_mock,
        patch("http.cookiejar.LWPCookieJar.save") as mock_save,
    ):
        mock_response = MagicMock()
        mock_response.status_code = 400
        mock_response.ok = False
        mock_response.json.return_value = {"error": "Bad Request"}
        mock_response.headers.get.return_value = "application/json"
        mock_request.return_value = mock_response
        pyicloud_session = PyiCloudSession(
            pyicloud_service_working, "", cookie_directory=""
        )
        with pytest.raises(PyiCloudAPIResponseException):
            pyicloud_session.request(
                "POST", "https://example.com", data={"key": "value"}
            )

        mock_request.assert_called_once_with(
            method="POST",
            url="https://example.com",
            data={"key": "value"},
            params=None,
            headers=None,
            cookies=None,
            files=None,
            auth=None,
            timeout=None,
            allow_redirects=True,
            proxies=None,
            hooks=None,
            stream=None,
            verify=None,
            cert=None,
            json=None,
        )
        mock_save.assert_called_once()
        assert open_mock.call_count == 2


def test_request_with_custom_headers(pyicloud_service_working: PyiCloudService) -> None:
    """Test the request method with custom headers."""
    with (
        patch("requests.Session.request") as mock_request,
        patch("builtins.open", new_callable=mock_open),
        patch("http.cookiejar.LWPCookieJar.save") as mock_save,
    ):
        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.json.return_value = {"data": "header test"}
        mock_response.headers.get.return_value = "application/json"
        mock_request.return_value = mock_response
        pyicloud_session = PyiCloudSession(
            pyicloud_service_working, "", cookie_directory=""
        )

        response: Response = pyicloud_session.request(
            "GET",
            "https://example.com",
            headers={"Custom-Header": "Value"},
        )
        assert response.json() == {"data": "header test"}
        assert response.headers.get("Content-Type") == "application/json"
        mock_request.assert_called_once_with(
            method="GET",
            url="https://example.com",
            data=None,
            headers={"Custom-Header": "Value"},
            params=None,
            cookies=None,
            files=None,
            auth=None,
            timeout=None,
            allow_redirects=True,
            proxies=None,
            hooks=None,
            stream=None,
            verify=None,
            cert=None,
            json=None,
        )
        mock_save.assert_called_once()


def test_request_error_handling_for_response_conditions() -> None:
    """Mock the get_webservice_url to return a valid fmip_url."""
    pyicloud_service = MagicMock(spec=PyiCloudService)
    with (
        pytest.raises(PyiCloudAPIResponseException),
        patch("requests.Session.request") as mock_request,
        patch("builtins.open", new_callable=mock_open),
        patch("os.path.exists", return_value=False),
        patch("http.cookiejar.LWPCookieJar.save"),
        patch.object(
            pyicloud_service,
            "get_webservice_url",
            return_value="https://fmip.example.com",
        ),
    ):
        # Mock the response with conditions that cause an error.
        mock_response = MagicMock()
        mock_response.status_code = 500
        mock_response.ok = False
        mock_response.json.return_value = {"error": "Server Error"}
        mock_response.headers.get.return_value = "application/json"
        mock_request.return_value = mock_response

        pyicloud_session = PyiCloudSession(pyicloud_service, "", cookie_directory="")
        pyicloud_service.data = {"session_token": "valid_token"}

        # Use the mocked fmip_url in the request.
        pyicloud_session.request("GET", "https://fmip.example.com/path")


def test_raise_error_2sa_required(pyicloud_session: PyiCloudSession) -> None:
    """Test the _raise_error method with a 2SA required exception."""
    with (
        pytest.raises(PyiCloud2SARequiredException),
        patch("pyicloud.base.PyiCloudService.requires_2sa", return_value=True),
    ):
        pyicloud_session._raise_error(
            401, reason="Missing X-APPLE-WEBAUTH-TOKEN cookie"
        )


def test_raise_error_service_not_activated(pyicloud_session: PyiCloudSession) -> None:
    """Test the _raise_error method with a service not activated exception."""
    with pytest.raises(PyiCloudServiceNotActivatedException):
        pyicloud_session._raise_error("ZONE_NOT_FOUND", reason="ServiceNotActivated")


def test_raise_error_access_denied(pyicloud_session: PyiCloudSession) -> None:
    """Test the _raise_error method with an access denied exception."""
    with pytest.raises(PyiCloudAPIResponseException):
        pyicloud_session._raise_error("ACCESS_DENIED", reason="ACCESS_DENIED")


def test_request_pcs_for_service_icdrs_not_disabled(
    pyicloud_service: PyiCloudService,
) -> None:
    """Test _request_pcs_for_service when ICDRS is not disabled (should early return)."""
    mock_logger = MagicMock()
    pyicloud_service.session = MagicMock()
    pyicloud_service.session.post = MagicMock(
        return_value=MagicMock(json=MagicMock(return_value={"isICDRSDisabled": False}))
    )
    pyicloud_service.params = {}
    with patch("pyicloud.base.LOGGER", mock_logger):
        pyicloud_service._send_pcs_request = MagicMock()
        pyicloud_service._request_pcs_for_service("photos")
        mock_logger.warning.assert_called_once_with("ICDRS is not disabled")
        pyicloud_service._send_pcs_request.assert_not_called()


def test_request_pcs_for_service_consent_needed_and_notification_sent(
    pyicloud_service: PyiCloudService,
) -> None:
    """Test _request_pcs_for_service when device consent is needed and notification is sent."""
    # First call: ICDRS disabled, device not consented
    # Second call: device consented (simulate after waiting)
    consent_states = [
        {"isICDRSDisabled": True, "isDeviceConsentedForPCS": False},
        {"isICDRSDisabled": True, "isDeviceConsentedForPCS": True},
    ]

    pyicloud_service._check_pcs_consent = MagicMock(side_effect=consent_states)
    pyicloud_service.session = MagicMock()
    pyicloud_service.params = {}
    pyicloud_service.session.post.return_value.json.side_effect = [
        {"isDeviceConsentNotificationSent": True},
        {"status": "success", "message": "ok"},
    ]
    with patch("time.sleep"):
        pyicloud_service._request_pcs_for_service("photos")
    pyicloud_service.session.post.assert_any_call(
        f"{pyicloud_service._setup_endpoint}/enableDeviceConsentForPCS",
        params=pyicloud_service.params,
    )
    # Should not raise


def test_request_pcs_for_service_consent_needed_and_notification_not_sent(
    pyicloud_service: PyiCloudService,
) -> None:
    """Test _request_pcs_for_service when device consent notification is not sent (should raise)."""
    pyicloud_service._check_pcs_consent = MagicMock(
        return_value={"isICDRSDisabled": True, "isDeviceConsentedForPCS": False}
    )
    pyicloud_service.session = MagicMock()
    pyicloud_service.params = {}
    pyicloud_service.session.post.return_value.json.return_value = {
        "isDeviceConsentNotificationSent": False
    }
    with pytest.raises(
        PyiCloudAPIResponseException, match="Unable to request PCS access!"
    ):
        pyicloud_service._request_pcs_for_service("photos")


def test_request_pcs_for_service_pcs_consent_waits(
    pyicloud_service: PyiCloudService,
) -> None:
    """Test _request_pcs_for_service waits for PCS consent and then proceeds."""
    # Simulate PCS consent not granted for first 2 tries, then granted
    consent_states = [
        {"isICDRSDisabled": True, "isDeviceConsentedForPCS": False},
        {"isICDRSDisabled": True, "isDeviceConsentedForPCS": False},
        {"isICDRSDisabled": True, "isDeviceConsentedForPCS": True},
    ]
    pyicloud_service._check_pcs_consent = MagicMock(side_effect=consent_states)
    pyicloud_service.session = MagicMock()
    pyicloud_service.params = {}
    pyicloud_service.session.post.return_value.json.return_value = {
        "isDeviceConsentNotificationSent": True
    }
    pyicloud_service._send_pcs_request = MagicMock(
        return_value={"status": "success", "message": "ok"}
    )
    with patch("time.sleep"):
        pyicloud_service._request_pcs_for_service("photos")
    assert pyicloud_service._send_pcs_request.called


def test_request_pcs_for_service_success_on_first_attempt(
    pyicloud_service: PyiCloudService,
) -> None:
    """Test _request_pcs_for_service grants PCS access on first attempt."""
    pyicloud_service._check_pcs_consent = MagicMock(
        return_value={"isICDRSDisabled": True, "isDeviceConsentedForPCS": True}
    )
    pyicloud_service.session = MagicMock()
    pyicloud_service.params = {}
    pyicloud_service._send_pcs_request = MagicMock(
        return_value={"status": "success", "message": "ok"}
    )
    pyicloud_service._request_pcs_for_service("photos")
    pyicloud_service._send_pcs_request.assert_called_once_with(
        "photos", derived_from_user_action=True
    )


def test_request_pcs_for_service_retries_on_cookie_messages(
    pyicloud_service: PyiCloudService,
) -> None:
    """Test _request_pcs_for_service retries on known cookie messages and succeeds."""
    pyicloud_service._check_pcs_consent = MagicMock(
        return_value={"isICDRSDisabled": True, "isDeviceConsentedForPCS": True}
    )
    pyicloud_service.session = MagicMock()
    pyicloud_service.params = {}
    responses = [
        {"status": "error", "message": "Requested the device to upload cookies."},
        {"status": "error", "message": "Cookies not available yet on server."},
        {"status": "success", "message": "ok"},
    ]
    pyicloud_service._send_pcs_request = MagicMock(side_effect=responses)
    with patch("time.sleep"):
        pyicloud_service._request_pcs_for_service("photos")
    assert pyicloud_service._send_pcs_request.call_count == 3


def test_request_pcs_for_service_raises_on_unknown_message(
    pyicloud_service: PyiCloudService,
) -> None:
    """Test _request_pcs_for_service raises on unknown PCS state message."""
    pyicloud_service._check_pcs_consent = MagicMock(
        return_value={"isICDRSDisabled": True, "isDeviceConsentedForPCS": True}
    )
    pyicloud_service.session = MagicMock()
    pyicloud_service.params = {}
    pyicloud_service._send_pcs_request = MagicMock(
        return_value={"status": "error", "message": "Some unknown error"}
    )
    mock_logger = MagicMock()

    with (
        pytest.raises(
            PyiCloudAPIResponseException, match="Unable to request PCS access!"
        ),
        patch("pyicloud.base.LOGGER", mock_logger),
    ):
        pyicloud_service._request_pcs_for_service("photos")
        mock_logger.error.assert_called()



"""Find my iPhone service."""

from typing import Any, Iterator, Optional

from requests import Response

from pyicloud.exceptions import PyiCloudNoDevicesException, PyiCloudServiceUnavailable
from pyicloud.services.base import BaseService
from pyicloud.session import PyiCloudSession


class FindMyiPhoneServiceManager(BaseService):
    """The 'Find my iPhone' iCloud service

    This connects to iCloud and return phone data including the near-realtime
    latitude and longitude.
    """

    def __init__(
        self,
        service_root: str,
        token_endpoint: str,
        session: PyiCloudSession,
        params: dict[str, Any],
        with_family=False,
    ) -> None:
        super().__init__(service_root, session, params)
        self.with_family: bool = with_family

        fmip_endpoint: str = f"{service_root}/fmipservice/client/web"
        self._fmip_refresh_url: str = f"{fmip_endpoint}/refreshClient"
        self._fmip_sound_url: str = f"{fmip_endpoint}/playSound"
        self._fmip_message_url: str = f"{fmip_endpoint}/sendMessage"
        self._fmip_lost_url: str = f"{fmip_endpoint}/lostDevice"
        self._fmip_erase_url: str = f"{fmip_endpoint}/remoteWipeWithUserAuth"
        self._erase_token_url: str = f"{token_endpoint}/fmipWebAuthenticate"

        self._devices: dict[str, AppleDevice] = {}
        self.refresh_client()

    def refresh_client(self) -> None:
        """Refreshes the FindMyiPhoneService endpoint,

        This ensures that the location data is up-to-date.

        """
        req: Response = self.session.post(
            self._fmip_refresh_url,
            params=self.params,
            json={
                "clientContext": {
                    "appName": "iCloud Find (Web)",
                    "appVersion": "2.0",
                    "apiVersion": "3.0",
                    "deviceListVersion": 1,
                    "fmly": self.with_family,
                }
            },
        )
        self.response: dict[str, Any] = req.json()

        for device_info in self.response["content"]:
            device_id: str = device_info["id"]
            if device_id not in self._devices:
                self._devices[device_id] = AppleDevice(
                    device_info,
                    self.params,
                    manager=self,
                    sound_url=self._fmip_sound_url,
                    lost_url=self._fmip_lost_url,
                    message_url=self._fmip_message_url,
                    erase_url=self._fmip_erase_url,
                    erase_token_url=self._erase_token_url,
                )
            else:
                self._devices[device_id].update(device_info)

        if not self._devices:
            raise PyiCloudNoDevicesException()

    def __getitem__(self, key) -> "AppleDevice":
        if isinstance(key, int):
            key = list(self.keys())[key]
        return self._devices[key]

    def __getattr__(self, attr: str) -> Any:
        return getattr(self._devices, attr)

    def __str__(self) -> str:
        return f"{self._devices}"

    def __repr__(self) -> str:
        return f"{self}"

    def __iter__(self) -> Iterator["AppleDevice"]:
        return iter(self._devices.values())

    def __len__(self) -> int:
        return len(self._devices)


class AppleDevice:
    """Apple device."""

    def __init__(
        self,
        content: dict[str, Any],
        params: dict[str, Any],
        manager: FindMyiPhoneServiceManager,
        sound_url: str,
        lost_url: str,
        erase_url: str,
        erase_token_url: str,
        message_url: str,
    ) -> None:
        self._content: dict[str, Any] = content
        self._manager: FindMyiPhoneServiceManager = manager
        self._params: dict[str, Any] = params

        self._sound_url: str = sound_url
        self._lost_url: str = lost_url
        self._erase_url: str = erase_url
        self._erase_token_url: str = erase_token_url
        self._message_url: str = message_url

    @property
    def session(self) -> PyiCloudSession:
        """Gets the session."""
        return self._manager.session

    def update(self, data) -> None:
        """Updates the device data."""
        self._content = data

    @property
    def location(self) -> Optional[dict[str, Any]]:
        """Updates the device location."""
        self._manager.refresh_client()
        return self._content["location"]

    def status(self, additional: Optional[list[str]] = None) -> dict[str, Any]:
        """Returns status information for device.

        This returns only a subset of possible properties.
        """
        self._manager.refresh_client()
        fields: list[str] = [
            "batteryLevel",
            "deviceDisplayName",
            "deviceStatus",
            "name",
        ]
        if additional is not None:
            fields += additional

        properties: dict[str, Any] = {}
        for field in fields:
            properties[field] = self._content.get(field)
        return properties

    def play_sound(self, subject="Find My iPhone Alert") -> None:
        """Send a request to the device to play a sound.

        It's possible to pass a custom message by changing the `subject`.
        """
        data: dict[str, Any] = {
            "device": self._content["id"],
            "subject": subject,
            "clientContext": {"fmly": True},
        }

        self.session.post(self._sound_url, params=self._params, json=data)

    def display_message(
        self, subject="Find My iPhone Alert", message="This is a note", sounds=False
    ) -> None:
        """Send a request to the device to play a sound.

        It's possible to pass a custom message by changing the `subject`.
        """
        data: dict[str, Any] = {
            "device": self._content["id"],
            "subject": subject,
            "sound": sounds,
            "userText": True,
            "text": message,
        }

        self.session.post(self._message_url, params=self._params, json=data)

    def lost_device(
        self,
        number: str,
        text: str = "This device has been lost. Please call me.",
        newpasscode: str = "",
    ) -> None:
        """Send a request to the device to trigger 'lost mode'.

        The device will show the message in `text`, and if a number has
        been passed, then the person holding the device can call
        the number without entering the passcode.
        """
        data: dict[str, Any] = {
            "text": text,
            "userText": True,
            "ownerNbr": number,
            "lostModeEnabled": True,
            "trackingEnabled": True,
            "device": self._content["id"],
            "passcode": newpasscode,
        }

        self.session.post(self._lost_url, params=self._params, json=data)

    def _get_erase_token(self) -> str:
        """Get the erase token for the Find My iPhone service."""
        data: dict[str, Any] = {
            "dsWebAuthToken": self.session.data.get("session_token"),
        }

        data = self.session.post(url=self._erase_token_url, json=data).json()
        if "tokens" not in data or "mmeFMIPWebEraseDeviceToken" not in data["tokens"]:
            raise PyiCloudServiceUnavailable("Find My iPhone erase token not available")
        return data["tokens"]["mmeFMIPWebEraseDeviceToken"]

    def erase_device(
        self,
        text: str = "This device has been lost. Please call me.",
        newpasscode: str = "",
    ) -> None:
        """Send a request to the device to start a remote erase."""
        data: dict[str, Any] = {
            "authToken": self._get_erase_token(),
            "text": text,
            "device": self._content["id"],
            "passcode": newpasscode,
        }

        self.session.post(self._erase_url, params=self._params, json=data)

    @property
    def data(self) -> dict[str, Any]:
        """Gets the device data."""
        return self._content

    def __getitem__(self, key) -> Any:
        return self._content[key]

    def __getattr__(self, attr) -> Any:
        if attr in self._content:
            return self._content[attr]
        raise AttributeError(
            f"'{type(self).__name__}' object has no attribute '{attr}'"
        )

    def __str__(self) -> str:
        return f"{self['deviceDisplayName']}: {self['name']}"

    def __repr__(self) -> str:
        return f"<AppleDevice({self})>"
